from enum import Enum
from inspect import isclass

from sqlalchemy.types import TypeDecorator, TypeEngine


__all__ = ["EnumType"]


class UndefinedEnumValue:
    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return f"<UndefinedEnumValue {self.value}>"


class EnumType(TypeDecorator):
    """Column type for storing Python :class:`Enum` types
    by converting them back-and-forth from & to a chosen SQL type.
    """

    cache_ok = True

    def __init__(self, enum, impl, allow_undefined_values=True):
        """Constructor.
        :type enum: Python :class:`Enum` subclass
        :param impl: Actual SQL type for the database side
        """
        if not issubclass(enum, Enum):
            raise TypeError("argument must be a Python enum class; got %r" % (enum,))
        if not (
            (isclass(impl) and issubclass(impl, TypeEngine))
            or isinstance(impl, TypeEngine)
        ):
            raise TypeError("argument must be an ORM type; got %r" % (impl,))

        self.enum = enum
        self.impl = impl
        self.allow_undefined_values = allow_undefined_values

    def copy(self):
        return EnumType(
            self.enum, self.impl, allow_undefined_values=self.allow_undefined_values
        )

    def load_dialect_impl(self, dialect):
        return dialect.type_descriptor(self.impl)

    def process_bind_param(self, value, dialect):
        """:type value: ``self.enum``"""
        if value is None:
            return None
        if not isinstance(value, self.enum):
            raise ValueError(
                "bind parameter is not of expected type %r, received %r"
                % (self.enum, value)
            )
        return value.value

    def process_result_value(self, value, dialect):
        """:type value: ``self.impl``"""
        if value is None:
            return None
        try:
            return self.enum(value)
        except ValueError:
            if self.allow_undefined_values:
                return UndefinedEnumValue(value)
            else:
                raise
